#include <KernelExport.h>
#include <sys/ioctl.h>
#include <SupportDefs.h>
#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <fs_interface.h>
#include <fs_info.h>

#include "3dofs.h"
#include "vnlist.h"


#define DRIVER_NAME "3do_fs"
#define ID DRIVER_NAME ": "

#define ROOT_VNID 0
#define BLOCK_SIZE 2048

#define DEBUG_DRIVER 0

#if DEBUG_DRIVER
#define DPRINTF(d,x) if (d) dprintf x
#else
#define DPRINTF(d,x) ((void)0)
#endif


// node information
typedef struct fs_node {
	vnode_id id;
	vnode_id parentid;
	mode_t mode;
	uint32 size;
	uint32 pdata;
	char name[33];
} fs_node;

// volume (per mount)
typedef struct fs_nspace {
	mount_id id;
	fs_node root;				// fs_node for "/"
	uint32 root_dir_head;
	int dev_fd;					// device file (/dev/disk/...)
	sem_id sem;
	struct dir_item *dir;
} fs_nspace;

// cookie for fs_*dir
typedef struct fs_dir_entry {
	char name[33];
	vnode_id id;
	struct fs_dir_entry *next;
} fs_dir_entry;

typedef struct fs_dircookie {
	struct fs_dir_entry *first;
	struct fs_dir_entry *current;
} fs_dircookie;


// proto types
void clean_up(fs_nspace *vol);
mode_t attr2mode(uint32 attr);
status_t read_dir(fs_nspace *vol, vnode_id id, struct dir_item **dir);


void clean_up(fs_nspace *vol)
{
	di_free(vol->dir);
	if (vol->dev_fd != -1) close(vol->dev_fd);
	if (vol->sem >= 0) delete_sem(vol->sem);
	free(vol);
}

// mount
#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_mount'. "
static status_t fs_mount(mount_id nsid, const char *devname, uint32 flags,
			const char *args, void **data, vnode_id *rootID)
{
	fs_nspace *vol;
	_3do_fs_superblock superblock;
	ssize_t size;
	sem_id sem;
	int e = B_OK;
	
	DPRINTF(DEBUG, (ID "fs_mount(%ld, %s, %ld, %s)\n", nsid, devname, flags, args));
	
	// vol create
	vol = (fs_nspace *)calloc(sizeof(fs_nspace), 1);
	if (vol == NULL) return ENOMEM;
	vol->sem = B_BAD_VALUE;
	
	vol->dev_fd = open(devname, O_RDONLY);
	if (vol->dev_fd == -1) {
		e = errno;
		DPRINTF(1, (ID "  " EFUNC "open() failed. %s\n", strerror(e)));
		clean_up(vol);
		return e;
	}
	
	// read superblock
	size = read(vol->dev_fd, &superblock, sizeof(superblock));
	if (size != sizeof(superblock)) {
		e = errno;
		DPRINTF(1, (ID "  " EFUNC "read() failed. %s\n", strerror(e)));
		clean_up(vol);
		return e;
	}
	
	if (memcmp("CD-ROM", superblock.magic_number, 6) != 0) {
		DPRINTF(1, (ID "  " EFUNC "Invalid magic number. (not 'CD-ROM')\n"));
		clean_up(vol);
		return EINVAL;
	}
	
	if ((sem = create_sem(1, "3dofs_sem")) < B_OK) {
		DPRINTF(1, (ID "  " EFUNC "create_sem() failed. %s\n", strerror(sem)));
		clean_up(vol);
		return sem;
	}
	
	vol->sem = sem;
	vol->root_dir_head = LENDIAN(superblock.root_pos) * BLOCK_SIZE;
	vol->root.id = ROOT_VNID;
	vol->root.parentid = vol->root.id;
	vol->root.mode = S_IFDIR | 0555;
	vol->id = nsid;
	*data = vol;
	*rootID = vol->root.id;
	
	// create "/" vnode
	if ((e = publish_vnode(nsid, *rootID, &vol->root)) != B_OK) {
		DPRINTF(1, (ID "  " EFUNC "publish_vnode() failed. %s\n", strerror(e)));
		clean_up(vol);
		return EINVAL;
	}
	
	DPRINTF(1, (ID "Mount successfully. vol_id = %ld\n", nsid));
	
	return B_OK;
}

// unmount
#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_unmount'. "
static status_t fs_unmount(void *_ns)
{
	fs_nspace *ns = _ns;
	DPRINTF(DEBUG, (ID "fs_unmount(%ld)\n", ns->id));
	clean_up(ns);
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_read_vnode'. "
static status_t fs_read_vnode(void *_ns, vnode_id vnid, void **node, bool reenter)
{
	fs_node *newNode;
	struct vnlist_item *entry;
	struct dir_item *dir;
	fs_nspace *vol = _ns;
	
	DPRINTF(DEBUG, (ID "fs_read_vnode(%ld, %lx, %d, *)\n", vol->id, (uint32)vnid, reenter));
	
	newNode = (fs_node *)calloc(sizeof(fs_node), 1);
	if (newNode == NULL) return ENOMEM;
	
	if (vnid == vol->root.id) {
		DPRINTF(1, (ID "  fs_read_vnode - root node requested.\n"));
		memcpy(newNode, &vol->root, sizeof(fs_node));
	} else {
		entry = vl_find_r(vol->dir, vnid, &dir);
		if (!entry) {
			free(newNode);
			return ENOENT;
		}
		
		newNode->id = vnid;
		newNode->parentid = dir->id;
		newNode->mode = entry->mode;
		newNode->size = entry->size;
		newNode->pdata = entry->pdata;
		strlcpy(newNode->name, entry->name, 33);
	}
	*node = newNode;
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_write_vnode'. "
static status_t fs_release_vnode(void *_ns, void *_node, bool reenter)
{
	fs_node *node = _node;
	
	DPRINTF(DEBUG, (ID "fs_write_vnode()\n"));
	
	if (node->id != ROOT_VNID) free(node);
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_walk'. "
static status_t fs_walk(void *_ns, void *_baseNode, const char *file, vnode_id *vnid, int *type)
{
	fs_node *newNode = NULL;
	struct dir_item *dir;
	struct vnlist_item *entry;
	fs_nspace *vol = _ns;
	fs_node *baseNode = _baseNode;
	
	DPRINTF(DEBUG, (ID "fs_walk(%ld, %lx, '%s')\n", vol->id, (int32)baseNode->id, file));
	
	if (strcmp(file, ".") == 0) {
		// current directory
		DPRINTF(DEBUG, (ID "  \".\" requested. vnid = %d\n", (int)baseNode->id));
		*vnid = baseNode->id;
		if (get_vnode(vol->id, *vnid, (void **)&newNode) != 0) return EINVAL;
		return B_OK;
	}
	if (strcmp(file, "..") == 0) {
		// parent directory
		DPRINTF(DEBUG, (ID "  \"..\" requested. vnid = %d\n", (int)baseNode->parentid));
		*vnid = baseNode->parentid;
		if (get_vnode(vol->id, *vnid, (void **)&newNode) != 0) return EINVAL;
		return B_OK;
	}
	
	dir = di_find_dir(vol->dir, baseNode->id);
	if (!dir) return ENOENT;
	
	entry = vl_find_name(dir->first, file);
	if (!entry) return ENOENT;
	
	*vnid = entry->id;
	if (get_vnode(vol->id, *vnid, (void **)&newNode) != 0) {
		return ENOENT;
	}
	newNode->parentid = baseNode->id;
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_get_vnode_name'. "
static status_t fs_get_vnode_name(void *_ns, void *_node, char *buffer, size_t bufferSize)
{
	fs_nspace *vol = _ns;
	fs_node *node = _node;
	
	DPRINTF(DEBUG, (ID "fs_get_vnode_name(%ld, %lx, *, %ld)\n", vol->id, (uint32)node->id, bufferSize));
	
	strlcpy(buffer, node->name, bufferSize);
	return B_OK;
}

mode_t attr2mode(uint32 attr)
{
	switch(LENDIAN(attr) & 0xf) {
		case SUB_DIRECTORY: return S_IFDIR | 0444;
		case PLAIN_FILE: return S_IFREG | 0444;
		default: return 0;
	}
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `read_dir'. "
status_t read_dir(fs_nspace *vol, vnode_id id, struct dir_item **dir)
{
	int e;
	uint32 head_pos, ncpy, attr;
	ssize_t size;
	struct _3do_fs_entry direntry;
	struct _3do_fs_dir_head *dir_head;
	struct dir_item *curdir;
	char buf[BLOCK_SIZE];
	char *p;
	
	// 既にキャッシュされていれば、それを使う
	*dir = di_find_dir(vol->dir, id);
	if (*dir) {
		DPRINTF(DEBUG, (ID "  found cache.\n"));
		return B_OK;
	}
	
	DPRINTF(DEBUG, (ID "  not found cache.\n"));
	
	curdir = di_add_dir(&vol->dir, id);
	if (curdir == NULL) return ENOMEM;
	*dir = curdir;
	
	if (id == ROOT_VNID) {
		head_pos = vol->root_dir_head;
	} else {
		if (lseek(vol->dev_fd, id, SEEK_SET) == -1) {
			e = errno;
			DPRINTF(1, (ID "  " EFUNC "lseek() failed. %s\n", strerror(e)));
			return e;
		}
		size = read(vol->dev_fd, &direntry, sizeof(struct _3do_fs_entry));
		if (size < sizeof(struct _3do_fs_entry)) {
			DPRINTF(1, (ID "  " EFUNC "read() failed.\n"));
			return EIO;
		}
		attr = LENDIAN(direntry.attr);
		if ((attr & 0xf) != SUB_DIRECTORY) {
			DPRINTF(1, (ID "  " EFUNC "not sub directory.\n"));
			return ENOENT;
		}
		head_pos = LENDIAN(direntry.start_block[0]) * BLOCK_SIZE;
	}
	
	while(true) {
		// 1ブロック読み込み
		if (lseek(vol->dev_fd, head_pos, SEEK_SET) == -1) {
			e = errno;
			DPRINTF(1, (ID "  " EFUNC "lseek() failed. %s\n", strerror(e)));
			return e;
		}
		size = read(vol->dev_fd, buf, sizeof(buf));
		if (size < sizeof(buf)) {
			DPRINTF(1, (ID "  " EFUNC "read() failed.\n"));
			return EIO;
		}
		p = buf;
		dir_head = (struct _3do_fs_dir_head *)p;
		p += sizeof(struct _3do_fs_dir_head);
		
		while(p < &buf[BLOCK_SIZE - sizeof(struct _3do_fs_entry)]) {
			struct _3do_fs_entry *entry = (struct _3do_fs_entry *)p;
			struct vnlist_item item;
			attr = LENDIAN(entry->attr) & 0xf;
			ncpy = LENDIAN(entry->num_of_copy_entries);
			
			item.mode = attr2mode(entry->attr);
			if (item.mode != 0) {
				item.id = head_pos + p - buf;			// ID は、エントリのディスク内の物理位置
				item.pdata = LENDIAN(entry->start_block[0]) * BLOCK_SIZE;
				item.size = LENDIAN(entry->size);
				strlcpy(item.name, entry->name, 33);
				item.name[32] = 0;
				vl_add_item(&curdir->first, &item);
			}
			
			if (attr & 0xff000000) break;
			p += sizeof(struct _3do_fs_entry) + (sizeof(uint32) * ncpy);
		}
		if (dir_head->cur_page == 0xffffffff) break;
		head_pos += BLOCK_SIZE;
	}
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_open_dir'. "
static status_t fs_open_dir(void *_ns, void *_node, void **_dir_cookie)
{
	int e;
	fs_dircookie *dir_cookie;
	struct dir_item *dir;
	struct vnlist_item *item;
	struct fs_dir_entry *fs_entry, *last = NULL, *cdir, *pdir;
	fs_nspace *vol = _ns;
	fs_node *node = _node;
	
	DPRINTF(DEBUG, (ID "fs_opendir(%ld, %lx)\n", vol->id, (int32)node->id));
	
	// enter critical section
	if (acquire_sem(vol->sem) != B_OK) {
		DPRINTF(1, (ID "  " EFUNC "acquire_sem() failed.\n"));
		return ETIMEDOUT;
	}
	
	e = read_dir(vol, node->id, &dir);
	
	// leave critical section
	release_sem(vol->sem);
	
	if (e != B_OK) return e;
	
	// create dircookie
	if ((dir_cookie = (fs_dircookie *)calloc(sizeof(fs_dircookie), 1)) == NULL) return ENOMEM;
	
	// "." & ".."
	cdir = (struct fs_dir_entry *)calloc(sizeof(struct fs_dir_entry), 1);
	strcpy(cdir->name, ".");
	cdir->id = node->id;
	
	pdir = (struct fs_dir_entry *)calloc(sizeof(struct fs_dir_entry), 1);
	strcpy(pdir->name, "..");
	pdir->id = node->parentid;
	
	cdir->next = pdir;
	dir_cookie->first = cdir;
	last = pdir;
	
	item = dir->first;
	while(item) {
		fs_entry = (struct fs_dir_entry *)calloc(sizeof(struct fs_dir_entry), 1);
		strlcpy(fs_entry->name, item->name, 33);
		fs_entry->id = item->id;
		last->next = fs_entry;
		last = fs_entry;
		item = item->next;
	}
	
	dir_cookie->current = dir_cookie->first;
	*_dir_cookie = dir_cookie;
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_close_dir'. "
static status_t fs_close_dir(void *_ns, void *_node, void *_dir_cookie)
{
	DPRINTF(DEBUG, (ID "fs_closedir()\n"));
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_free_dircookie'. "
static status_t fs_free_dir_cookie(void *_ns, void *_node, void *_dir_cookie)
{
	fs_dircookie *dir_cookie = _dir_cookie;
	struct fs_dir_entry *fs_entry = dir_cookie->first, *next;
	
	DPRINTF(DEBUG, (ID "fs_free_dircookie()\n"));
	
	while(fs_entry) {
		next = fs_entry->next;
		free(fs_entry);
		fs_entry = next;
	}
	
	free(dir_cookie);
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_rewind_dir'. "
static status_t fs_rewind_dir(void *_ns, void *_node, void *_dir_cookie)
{
	fs_dircookie *dir_cookie = _dir_cookie;
	DPRINTF(DEBUG, (ID "fs_rewinddir()\n"));
	dir_cookie->current = dir_cookie->first;
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_read_dir'. "
static status_t fs_read_dir(void *_ns, void *_node, void *_dir_cookie, struct dirent *entry,
	size_t bufsize, uint32 *num)
{
	struct fs_dir_entry *fs_entry;
	uint32 namelen;
	fs_dircookie *dir_cookie = _dir_cookie;
	fs_nspace *vol = _ns;
	DPRINTF(DEBUG, (ID "fs_readdir()\n"));
	
	if (dir_cookie->current == NULL) {
		// no more entries
		*num = 0;
		DPRINTF(DEBUG, (ID "  no more entries.\n"));
	} else {
		fs_entry = dir_cookie->current;
		entry->d_dev = vol->id;
		entry->d_ino = fs_entry->id;
		namelen = strlen(fs_entry->name);
		entry->d_reclen = 2 * (sizeof(dev_t) + sizeof(ino_t)) + sizeof(unsigned short) + namelen + 1;
		strcpy(entry->d_name, fs_entry->name);
		dir_cookie->current = fs_entry->next;
		*num = 1;
		DPRINTF(DEBUG, (ID "  '%s'\n", entry->d_name));
	}
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_open'. "
static status_t fs_open(void *_ns, void *_node, int omode, void **cookie)
{
	fs_nspace *vol = _ns;
	fs_node *node = _node;
	
	DPRINTF(DEBUG, (ID "fs_open(%ld, %lx, %d, %p)\n", vol->id, (int32)node->id, omode, cookie));
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_read'. "
static status_t fs_read(void *_ns, void *_node, void *cookie, off_t pos, void *buf, size_t *len)
{
	int e;
	off_t ppos;
	ssize_t size;
	size_t rlen;
	fs_nspace *vol = _ns;
	fs_node *node = _node;
	
	DPRINTF(DEBUG, (ID "fs_read(%ld, %lx, %p, %ld, %ld)\n", vol->id, (uint32)node->id, cookie, (uint32)pos, (uint32)*len));
	
	if ((pos < 0) || (*len < 0)) return EINVAL;
	if (node->pdata == 0) {
		*len = 0;
		return B_OK;
	}
	
	if (pos >= node->size) {
		*len = 0;
		return B_OK;
	}
	
	// 読み出し開始位置
	ppos = node->pdata + pos;
	
	// 読み出し量
	rlen = min(node->size - pos, *len);
	
	// enter critical section
	if (acquire_sem(vol->sem) != B_OK) {
		DPRINTF(1, (ID "  " EFUNC "acquire_sem() failed.\n"));
		return ETIMEDOUT;
	}
	if (lseek(vol->dev_fd, ppos, SEEK_SET) == -1) {
		e = errno;
		DPRINTF(1, (ID "  " EFUNC "lseek() failed. %s\n", strerror(e)));
		release_sem(vol->sem);
		return e;
	}
	size = read(vol->dev_fd, buf, rlen);
	if (size == -1) {
		e = errno;
		DPRINTF(1, (ID "  " EFUNC "read() failed. %s\n", strerror(e)));
		release_sem(vol->sem);
		return EIO;
	}
	
	// leave critical section
	release_sem(vol->sem);
	
	*len = size;
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_close'. "
static status_t fs_close(void *_ns, void *_node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_close()\n"));
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_free_cookie'. "
static status_t fs_free_cookie(void *_ns, void *_node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_free_cookie()\n"));
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_access'. "
static status_t fs_access(void *ns, void *node, int mode)
{
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_read_stat'. "
static status_t fs_read_stat(void *_ns, void *_node, struct stat *st)
{
	fs_nspace *vol = _ns;
	fs_node *node = _node;
	
	DPRINTF(DEBUG, (ID "fs_rstat(%ld, %lx)\n", vol->id, (int32)node->id));
	
	st->st_dev = vol->id;
	st->st_ino = node->id;
	st->st_nlink = 1;
	st->st_uid = 0;
	st->st_gid = 0;
	st->st_blksize = 65536;
	st->st_atime = 0;
	st->st_mtime = 0;
	st->st_ctime = 0;
	st->st_crtime = 0;
	st->st_mode = node->mode;
	st->st_size = node->size;
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
#undef EFUNC
#define EFUNC "Error: In function `fs_read_fs_stat'. "
static status_t fs_read_fs_stat(void *_ns, struct fs_info *info)
{
	DPRINTF(DEBUG, (ID "fs_rfsstat()\n"));
	
	info->flags = 0
		 | B_FS_IS_READONLY
		 | B_FS_IS_REMOVABLE
		 | B_FS_IS_PERSISTENT
	;
	info->block_size = BLOCK_SIZE;
	info->io_size = 65536;
	info->total_blocks = 0;
	info->free_blocks = 0;
	strcpy(info->device_name, "");
	strcpy(info->volume_name, "3dofs");
	strcpy(info->fsh_name, "3dofs");
	
	return B_OK;
}


// Scanning
/*
static float fs_identify_partition(int fd, partition_data *partition, void **_cookie)
{
	iso9660_info info;
	status_t status = iso9660_fs_identify(fd, &info);
	if (status != B_OK) return status;
	
	identify_cookie *cookie = new identify_cookie;
	memcpy(&cookie->info, &info, sizeof(info));
	
	*_cookie = cookie;
	return 0.8f;
}


static status_t fs_scan_partition(int fd, partition_data *partition, void *_cookie)
{
	identify_cookie *cookie = (identify_cookie *)_cookie;

	partition->status = B_PARTITION_VALID;
	partition->flags |= B_PARTITION_FILE_SYSTEM | B_PARTITION_READ_ONLY ;
	partition->block_size = ISO_PVD_SIZE;
	partition->content_size = ISO_PVD_SIZE * cookie->info.maxBlocks;
	partition->content_name = strdup(cookie->info.get_preferred_volume_name());
	if (partition->content_name == NULL) return ENOMEM;
	return B_OK;
}


static void fs_free_identify_partition_cookie(partition_data *partition, void *_cookie)
{
	identify_cookie *cookie = (identify_cookie *)_cookie;
	delete cookie;
}
*/

//

static status_t _3do_std_ops(int32 op, ...)
{
	switch (op) {
		case B_MODULE_INIT:
			return B_OK;
		case B_MODULE_UNINIT:
			return B_OK;
		default:
			return B_ERROR;
	}
}

static file_system_module_info s3DOFileSystem = {
	{
		"file_systems/3dofs" B_CURRENT_FS_API_VERSION,
		0,
		_3do_std_ops,
	},
	
	"3DO File System",
	
	// scanning
	NULL,   // fs_identify_partition,
	NULL,   // fs_scan_partition,
	NULL,   // fs_free_identify_partition_cookie,
	NULL,	// free_partition_content_cookie()
	
	&fs_mount,
	&fs_unmount,
	&fs_read_fs_stat,
	NULL,
	NULL,
	
	/* vnode operations */
	&fs_walk,
	&fs_get_vnode_name,
	&fs_read_vnode,
	&fs_release_vnode,
	NULL, 	// &fs_remove_vnode()
	
	/* VM file access */
	NULL, 	// &fs_can_page
	NULL,	// &fs_read_pages
	NULL, 	// &fs_write_pages
	
	NULL,	// &fs_get_file_map
	
	NULL, 	// &fs_ioctl
	NULL, 	// &fs_set_flags
	NULL,	// &fs_select
	NULL,	// &fs_deselect
	NULL, 	// &fs_fsync
	
	NULL,	// &fs_read_link,
	NULL,	// write link
	NULL, 	// &fs_create_symlink,
	
	NULL, 	// &fs_link,
	NULL,	// &fs_unlink
	NULL, 	// &fs_rename
	
	&fs_access,
	&fs_read_stat,
	NULL, 	// &fs_write_stat
	
	/* file operations */
	NULL, 	// &fs_create
	&fs_open,
	&fs_close,
	&fs_free_cookie,
	&fs_read,
	NULL, 	// &fs_write
	
	/* directory operations */
	NULL, 	// &fs_create_dir
	NULL, 	// &fs_remove_dir
	&fs_open_dir,
	&fs_close_dir,
	&fs_free_dir_cookie,
	&fs_read_dir,
	&fs_rewind_dir,
	
	/* attribute directory operations */
	NULL, 	// &fs_open_attr_dir
	NULL, 	// &fs_close_attr_dir
	NULL,	// &fs_free_attr_dir_cookie
	NULL,	// &fs_read_attr_dir
	NULL,	// &fs_rewind_attr_dir
	
	/* attribute operations */
	NULL,	// &fs_create_attr
	NULL, 	// &fs_open_attr
	NULL,	// &fs_close_attr
	NULL,	// &fs_free_attr_cookie
	NULL,	// &fs_read_attr
	NULL,	// &fs_write_attr
	
	NULL,	// &fs_read_attr_stat
	NULL,	// &fs_write_attr_stat
	NULL,	// &fs_rename_attr
	NULL,	// &fs_remove_attr
	
	/* index directory & index operations */
	NULL,	// &fs_open_index_dir
	NULL,	// &fs_close_index_dir
	NULL,	// &fs_free_index_dir_cookie
	NULL,	// &fs_read_index_dir
	NULL,	// &fs_rewind_index_dir
	
	NULL,	// &fs_create_index
	NULL,	// &fs_remove_index
	NULL,	// &fs_stat_index
	
	/* query operations */
	NULL,	// &fs_open_query
	NULL,	// &fs_close_query
	NULL,	// &fs_free_query_cookie
	NULL,	// &fs_read_query
	NULL,	// &fs_rewind_query
};

module_info *modules[] = {
	(module_info *)&s3DOFileSystem,
	NULL,
};
